# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.


# This is testing the function load feature for the single event annotation
# The event shouldn't be int
def main(event: int) -> str:
    return 'BAD'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from azure.functions import meta


class Binding(meta.InConverter, meta.OutConverter,
              binding='fooType'):
    pass



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from azure_functions_worker import protos
from azure_functions_worker import testutils


class TestGenericFunctions(testutils.AsyncTestCase):
    generic_funcs_dir = testutils.UNIT_TESTS_FOLDER / 'generic_functions'

    async def test_mock_generic_as_str(self):
        async with testutils.start_mockhost(
                script_root=self.generic_funcs_dir) as host:

            func_id, r = await host.load_function('foobar_as_str')

            self.assertEqual(r.response.function_id, func_id)
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)

            _, r = await host.invoke_function(
                'foobar_as_str', [
                    protos.ParameterBinding(
                        name='input',
                        data=protos.TypedData(
                            string='test'
                        )
                    )
                ]
            )
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)
            self.assertEqual(
                r.response.return_value,
                protos.TypedData(string='test')
            )

    async def test_mock_generic_as_bytes(self):
        async with testutils.start_mockhost(
                script_root=self.generic_funcs_dir) as host:

            func_id, r = await host.load_function('foobar_as_bytes')

            self.assertEqual(r.response.function_id, func_id)
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)

            _, r = await host.invoke_function(
                'foobar_as_bytes', [
                    protos.ParameterBinding(
                        name='input',
                        data=protos.TypedData(
                            bytes=b'\x00\x01'
                        )
                    )
                ]
            )
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)
            self.assertEqual(
                r.response.return_value,
                protos.TypedData(bytes=b'\x00\x01')
            )

    async def test_mock_generic_as_str_no_anno(self):
        async with testutils.start_mockhost(
                script_root=self.generic_funcs_dir) as host:

            func_id, r = await host.load_function('foobar_as_str_no_anno')

            self.assertEqual(r.response.function_id, func_id)
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)

            _, r = await host.invoke_function(
                'foobar_as_str_no_anno', [
                    protos.ParameterBinding(
                        name='input',
                        data=protos.TypedData(
                            string='test'
                        )
                    )
                ]
            )
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)
            self.assertEqual(
                r.response.return_value,
                protos.TypedData(string='test')
            )

    async def test_mock_generic_as_bytes_no_anno(self):
        async with testutils.start_mockhost(
                script_root=self.generic_funcs_dir) as host:

            func_id, r = await host.load_function('foobar_as_bytes_no_anno')

            self.assertEqual(r.response.function_id, func_id)
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)

            _, r = await host.invoke_function(
                'foobar_as_bytes_no_anno', [
                    protos.ParameterBinding(
                        name='input',
                        data=protos.TypedData(
                            bytes=b'\x00\x01'
                        )
                    )
                ]
            )
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)
            self.assertEqual(
                r.response.return_value,
                protos.TypedData(bytes=b'\x00\x01')
            )

    async def test_mock_generic_should_not_support_implicit_output(self):
        async with testutils.start_mockhost(
                script_root=self.generic_funcs_dir) as host:

            func_id, r = await host.load_function('foobar_implicit_output')

            self.assertEqual(r.response.function_id, func_id)
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)

            _, r = await host.invoke_function(
                'foobar_as_bytes_no_anno', [
                    protos.ParameterBinding(
                        name='input',
                        data=protos.TypedData(
                            bytes=b'\x00\x01'
                        )
                    )
                ]
            )
            # It should fail here, since generic binding requires
            # $return statement in function.json to pass output
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Failure)

    async def test_mock_generic_should_support_without_datatype(self):
        async with testutils.start_mockhost(
                script_root=self.generic_funcs_dir) as host:

            func_id, r = await host.load_function('foobar_with_no_datatype')

            self.assertEqual(r.response.function_id, func_id)
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)

            _, r = await host.invoke_function(
                'foobar_with_no_datatype', [
                    protos.ParameterBinding(
                        name='input',
                        data=protos.TypedData(
                            bytes=b'\x00\x01'
                        )
                    )
                ]
            )
            # It should fail here, since the generic binding requires datatype
            # to be defined in function.json
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Failure)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import asyncio
import azure.functions


async def main(req: azure.functions.HttpRequest, context):
    await asyncio.sleep(0.1)
    return 'Hello Async World!'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import sys
import logging
import asyncio

import azure.functions as func


logger = logging.getLogger('custom_logger')

# Attempt to log info into system log from customer code
disguised_logger = logging.getLogger('azure_functions_worker')


async def parallelly_print():
    await asyncio.sleep(0.1)
    print('parallelly_print')


async def parallelly_log_info():
    await asyncio.sleep(0.2)
    logging.info('parallelly_log_info at root logger')


async def parallelly_log_warning():
    await asyncio.sleep(0.3)
    logging.warning('parallelly_log_warning at root logger')


async def parallelly_log_error():
    await asyncio.sleep(0.4)
    logging.error('parallelly_log_error at root logger')


async def parallelly_log_exception():
    await asyncio.sleep(0.5)
    try:
        raise Exception('custom exception')
    except Exception:
        logging.exception('parallelly_log_exception at root logger',
                          exc_info=sys.exc_info())


async def parallelly_log_custom():
    await asyncio.sleep(0.6)
    logger.info('parallelly_log_custom at custom_logger')


async def parallelly_log_system():
    await asyncio.sleep(0.7)
    disguised_logger.info('parallelly_log_system at disguised_logger')


async def main(req: func.HttpRequest) -> func.HttpResponse:
    loop = asyncio.get_event_loop()

    # Create multiple tasks and schedule it into one asyncio.wait blocker
    task_print: asyncio.Task = loop.create_task(parallelly_print())
    task_info: asyncio.Task = loop.create_task(parallelly_log_info())
    task_warning: asyncio.Task = loop.create_task(parallelly_log_warning())
    task_error: asyncio.Task = loop.create_task(parallelly_log_error())
    task_exception: asyncio.Task = loop.create_task(parallelly_log_exception())
    task_custom: asyncio.Task = loop.create_task(parallelly_log_custom())
    task_disguise: asyncio.Task = loop.create_task(parallelly_log_system())

    # Create an awaitable future and occupy the current event loop resource
    future = loop.create_future()
    loop.call_soon_threadsafe(future.set_result, 'callsoon_log')

    # WaitAll
    await asyncio.wait([task_print, task_info, task_warning, task_error,
                        task_exception, task_custom, task_disguise, future])

    # Log asyncio low-level future result
    logging.info(future.result())

    return 'OK-hijack-current-event-loop'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import json
import azure.functions


def main(req: azure.functions.HttpRequest) -> str:
    return json.dumps(dict(req.route_params))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as azf


def main(req: azf.HttpRequest):
    return azf.HttpResponse('<h1>Hello World™</h1>',
                            mimetype='text/html')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as func


# Retrieve the event data from storage blob and return it as Http response
def main(req: func.HttpRequest, file: func.InputStream) -> str:
    return func.HttpResponse(body=file.read().decode('utf-8'),
                             status_code=200,
                             mimetype='application/json')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import sys
import azure.functions


def main(req: azure.functions.HttpRequest):
    flush_required = False
    is_console_log = False
    is_stderr = False
    message = req.params.get('message', '')

    if req.params.get('flush') == 'true':
        flush_required = True
    if req.params.get('console') == 'true':
        is_console_log = True
    if req.params.get('is_stderr') == 'true':
        is_stderr = True

    # Adding LanguageWorkerConsoleLog will make function host to treat
    # this as system log and will be propagated to kusto
    prefix = 'LanguageWorkerConsoleLog' if is_console_log else ''
    print(f'{prefix} {message}'.strip(),
          file=sys.stderr if is_stderr else sys.stdout,
          flush=flush_required)

    return 'OK-print-logging'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from typing import Dict


def main(input: Dict[str, str]) -> Dict[str, str]:
    result = input.copy()
    if result.get('bird'):
        result['bird'] = result['bird'][::-1]

    return result



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import Optional
import logging
import logging.handlers
import sys

from .constants import CONSOLE_LOG_PREFIX


logger: logging.Logger = logging.getLogger('azure_functions_worker')
error_logger: logging.Logger = (
    logging.getLogger('azure_functions_worker_errors'))

handler: Optional[logging.Handler] = None
error_handler: Optional[logging.Handler] = None


def setup(log_level, log_destination):
    # Since handler and error_handler are moved to the global scope,
    # before assigning to these handlers, we should define 'global' keyword
    global handler
    global error_handler

    if log_level == 'TRACE':
        log_level = 'DEBUG'

    formatter = logging.Formatter(f'{CONSOLE_LOG_PREFIX}'
                                  ' %(levelname)s: %(message)s')

    if log_destination is None:
        # With no explicit log destination we do split logging,
        # errors go into stderr, everything else -- to stdout.
        error_handler = logging.StreamHandler(sys.stderr)
        error_handler.setFormatter(formatter)
        error_handler.setLevel(getattr(logging, log_level))

        handler = logging.StreamHandler(sys.stdout)

    elif log_destination in ('stdout', 'stderr'):
        handler = logging.StreamHandler(getattr(sys, log_destination))

    elif log_destination == 'syslog':
        handler = logging.handlers.SysLogHandler()

    else:
        handler = logging.FileHandler(log_destination)

    if error_handler is None:
        error_handler = handler

    handler.setFormatter(formatter)
    handler.setLevel(getattr(logging, log_level))

    logger.addHandler(handler)
    logger.setLevel(getattr(logging, log_level))

    error_logger.addHandler(error_handler)
    error_logger.setLevel(getattr(logging, log_level))


def disable_console_logging() -> None:
    # We should only remove the sys.stdout stream, as error_logger is used for
    # unexpected critical error logs handling.
    if logger and handler:
        handler.flush()
        logger.removeHandler(handler)


def enable_console_logging() -> None:
    if logger and handler:
        logger.addHandler(handler)


def is_system_log_category(ctg: str) -> bool:
    # Category starts with 'azure_functions_worker' or
    # 'azure_functions_worker_errors' will be treated as system logs
    return ctg.lower().startswith('azure_functions_worker')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as func


# An HttpTrigger to generating EventHub event from EventHub Output Binding
def main(req: func.HttpRequest, event: func.Out[str]):
    event.set(req.get_body().decode('utf-8'))

    return 'OK'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

# Both customer code and third-party package has the same name pytest.
# When using absolute import, should pick customer's package.
import __app__.pytest as pt


def main(req) -> str:
    return f'pt.__version__ = {pt.__version__}'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

# Import a module from thirdparty package azure-eventhub
import azure.eventhub as eh


def main(req) -> str:
    return f'eh = {eh.__name__}'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

# Prefixes
CONSOLE_LOG_PREFIX = "LanguageWorkerConsoleLog"

# Capabilities
RAW_HTTP_BODY_BYTES = "RawHttpBodyBytes"
TYPED_DATA_COLLECTION = "TypedDataCollection"
RPC_HTTP_BODY_ONLY = "RpcHttpBodyOnly"
RPC_HTTP_TRIGGER_METADATA_REMOVED = "RpcHttpTriggerMetadataRemoved"

# Debug Flags
PYAZURE_WEBHOST_DEBUG = "PYAZURE_WEBHOST_DEBUG"

# Python Specific Feature Flags and App Settings
PYTHON_ROLLBACK_CWD_PATH = "PYTHON_ROLLBACK_CWD_PATH"
PYTHON_THREADPOOL_THREAD_COUNT = "PYTHON_THREADPOOL_THREAD_COUNT"

# Setting Defaults
PYTHON_THREADPOOL_THREAD_COUNT_DEFAULT = 1
PYTHON_THREADPOOL_THREAD_COUNT_MIN = 1
PYTHON_THREADPOOL_THREAD_COUNT_MAX = 32

# External Site URLs
MODULE_NOT_FOUND_TS_URL = "https://aka.ms/functions-modulenotfound"



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# import azure.durable_functions as df


def generator_function(context):
    final_result = yield context.call_activity('activity_trigger', 'foobar')
    return final_result


def main(context):
    # orchestrate = df.Orchestrator.create(generator_function)
    # result = orchestrate(context)
    # return result
    return f'{context} :)'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as azf


def main(timer: azf.TimerRequest, pastdue: azf.Out[str]):
    pastdue.set(str(timer.past_due))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as azf


def main(req: azf.HttpRequest, file: azf.Out[str]) -> str:
    file.set(req.get_body())
    return 'OK'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as func


def main(req: func.HttpRequest, file: func.InputStream) -> str:
    return func.HttpResponse(
        file.read().decode('utf-8'), mimetype='application/json')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import typing
import azure.functions as azf


def main(req: azf.HttpRequest, msgs: azf.Out[typing.List[str]]):
    msgs.set(['one', 'two'])



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import json
import time
from datetime import datetime
from dateutil import parser, tz

from azure_functions_worker import testutils


class TestEventHubFunctions(testutils.WebHostTestCase):
    """Test EventHub Trigger and Output Bindings (cardinality: one).

    Each testcase consists of 3 part:
    1. An eventhub_output HTTP trigger for generating EventHub event
    2. An actual eventhub_trigger EventHub trigger for storing event into blob
    3. A get_eventhub_triggered HTTP trigger for retrieving event info blob
    """

    @classmethod
    def get_script_dir(cls):
        return testutils.E2E_TESTS_FOLDER / 'eventhub_functions'

    @testutils.retryable_test(3, 5)
    def test_eventhub_trigger(self):
        # Generate a unique event body for the EventHub event
        data = str(round(time.time()))
        doc = {'id': data}

        # Invoke eventhub_output HttpTrigger to generate an Eventhub Event.
        r = self.webhost.request('POST', 'eventhub_output',
                                 data=json.dumps(doc))
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.text, 'OK')

        # Once the event get generated, allow function host to pool from
        # EventHub and wait for eventhub_trigger to execute,
        # converting the event metadata into a blob.
        time.sleep(5)

        # Call get_eventhub_triggered to retrieve event metadata from blob.
        r = self.webhost.request('GET', 'get_eventhub_triggered')
        self.assertEqual(r.status_code, 200)
        response = r.json()

        # Check if the event body matches the initial data
        self.assertEqual(response, doc)

    @testutils.retryable_test(3, 5)
    def test_eventhub_trigger_with_metadata(self):
        # Generate a unique event body for EventHub event
        # Record the start_time and end_time for checking event enqueue time
        start_time = datetime.now(tz=tz.UTC)
        random_number = str(round(time.time()) % 1000)
        req_body = {
            'body': random_number
        }

        # Invoke metadata_output HttpTrigger to generate an EventHub event
        # from azure-eventhub SDK
        r = self.webhost.request('POST', 'metadata_output',
                                 data=json.dumps(req_body))
        self.assertEqual(r.status_code, 200)
        self.assertIn('OK', r.text)
        end_time = datetime.now(tz=tz.UTC)

        # Once the event get generated, allow function host to pool from
        # EventHub and wait for eventhub_trigger to execute,
        # converting the event metadata into a blob.
        time.sleep(5)

        # Call get_metadata_triggered to retrieve event metadata from blob
        r = self.webhost.request('GET', 'get_metadata_triggered')
        self.assertEqual(r.status_code, 200)

        # Check if the event body matches the unique random_number
        event = r.json()
        self.assertEqual(event['body'], random_number)

        # EventhubEvent property check
        # Reenable these lines after enqueued_time property is fixed
        # enqueued_time = parser.isoparse(event['enqueued_time'])
        # self.assertIsNotNone(enqueued_time)
        self.assertIsNone(event['partition_key'])  # There's only 1 partition
        self.assertGreaterEqual(event['sequence_number'], 0)
        self.assertIsNotNone(event['offset'])

        # Check if the event contains proper metadata fields
        self.assertIsNotNone(event['metadata'])
        metadata = event['metadata']
        sys_props = metadata['SystemProperties']
        enqueued_time = parser.isoparse(metadata['EnqueuedTimeUtc'])
        self.assertTrue(start_time < enqueued_time < end_time)
        self.assertIsNone(sys_props['PartitionKey'])
        self.assertGreaterEqual(sys_props['SequenceNumber'], 0)
        self.assertIsNotNone(sys_props['Offset'])

        # Metadata check eventhub trigger name
        self.assertEqual(metadata['sys']['MethodName'], 'metadata_trigger')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import json

import azure.functions as func


def main(event: func.EventHubEvent) -> str:
    return json.dumps(event.iothub_metadata)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from urllib.request import urlopen

import azure.functions as func


def main(req: func.HttpRequest) -> str:
    image_url = req.params.get('img')
    urlopen(image_url).read()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as func


# An HttpTrigger to generating EventHub event from EventHub Output Binding
def main(req: func.HttpRequest) -> str:
    events = req.get_body().decode('utf-8')
    return events



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.


def main(req):
    # This function will fail, as we don't auto-convert "bytes" to "http".
    return b'Hello World!'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as azf


def main(req: azf.HttpRequest, file: bytes) -> str:
    assert isinstance(file, bytes)
    return file.decode('utf-8')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# Import simple module with implicit statement should now be acceptable
import notfound


def modulenotfound(req) -> str:
    return f'notfound = {notfound.__name__}'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
"""Main entrypoint."""


import argparse

from . import dispatcher
from . import logging
from ._thirdparty import aio_compat
from .logging import error_logger, logger


def parse_args():
    parser = argparse.ArgumentParser(
        description='Python Azure Functions Worker')
    parser.add_argument('--host',
                        help="host address")
    parser.add_argument('--port', type=int,
                        help='id for the requests')
    parser.add_argument('--workerId', dest='worker_id',
                        help='id for the worker')
    parser.add_argument('--requestId', dest='request_id',
                        help='log destination: stdout, stderr, '
                             'syslog, or a file path')
    parser.add_argument('--log-level', type=str, default='INFO',
                        choices=['TRACE', 'INFO', 'WARNING', 'ERROR'],
                        help="log level: 'TRACE', 'INFO', 'WARNING', "
                             "or 'ERROR'")
    parser.add_argument('--log-to', type=str, default=None,
                        help='log destination: stdout, stderr, '
                             'syslog, or a file path')
    parser.add_argument('--grpcMaxMessageLength', type=int,
                        dest='grpc_max_msg_len')
    return parser.parse_args()


def main():
    args = parse_args()
    logging.setup(log_level=args.log_level, log_destination=args.log_to)

    logger.info('Starting Azure Functions Python Worker.')
    logger.info('Worker ID: %s, Request ID: %s, Host Address: %s:%s',
                args.worker_id, args.request_id, args.host, args.port)

    try:
        return aio_compat.run(start_async(
            args.host, args.port, args.worker_id, args.request_id))
    except Exception:
        error_logger.exception('unhandled error in functions worker')
        raise


async def start_async(host, port, worker_id, request_id):
    disp = await dispatcher.Dispatcher.connect(host=host, port=port,
                                               worker_id=worker_id,
                                               request_id=request_id,
                                               connect_timeout=5.0)

    await disp.dispatch_forever()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import logging

import azure.functions
import does_not_exist  # Noqa


logger = logging.getLogger('my function')


def main(req: azure.functions.HttpRequest):
    logger.info('Function should fail before hitting main')
    return 'OK-async'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.


"""This module pytest is provided inside customer's code,
used for checking module name collision"""
__version__ = 'from.customer.code'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as azf


def main(req: azf.HttpRequest, file: str) -> str:
    assert isinstance(file, str)
    return file



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as func


# Retrieve the event data from storage blob and return it as Http response
def main(req: func.HttpRequest, testEntities):
    return func.HttpResponse(status_code=200, body=testEntities)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import logging

import azure.functions


logger = logging.getLogger('my function')


def main(req: azure.functions.HttpRequest):
    logger.info('logging info', exc_info=True)
    logger.warning('logging warning', exc_info=True)
    logger.debug('logging debug', exc_info=True)
    logger.error('logging error', exc_info=True)
    return 'OK-user-debug'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from datetime import datetime

import azure.functions as func


def main(req: func.HttpRequest,
         outputEvent: func.Out[func.EventGridOutputEvent]) -> func.HttpResponse:
    test_uuid = req.params.get('test_uuid')
    data_to_event_grid = func.EventGridOutputEvent(id="test-id",
                                                   data={
                                                       "test_uuid": test_uuid
                                                   },
                                                   subject="test-subject",
                                                   event_type="test-event-1",
                                                   event_time=datetime.utcnow(),
                                                   data_version="1.0")

    outputEvent.set(data_to_event_grid)
    r_value = "Sent event with subject: {}, id: {}, data: {}, event_type: {} " \
              "to EventGrid!".format(data_to_event_grid.subject,
                                     data_to_event_grid.id,
                                     data_to_event_grid.get_json(),
                                     data_to_event_grid.event_type)
    return func.HttpResponse(r_value)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as azf


def main(req: azf.HttpRequest, foo: azf.Out[azf.HttpResponse]):
    foo.set(azf.HttpResponse(body='hello', status_code=201))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import time
import unittest
import uuid

import requests

from azure_functions_worker import testutils


class TestEventGridFunctions(testutils.WebHostTestCase):

    @classmethod
    def get_script_dir(cls):
        return testutils.E2E_TESTS_FOLDER / 'eventgrid_functions'

    def eventgrid_webhook_request(self, meth, funcname, *args, **kwargs):
        request_method = getattr(requests, meth.lower())
        url = f'{self.webhost._addr}/runtime/webhooks/eventgrid'
        params = dict(kwargs.pop('params', {}))
        params['functionName'] = funcname
        if 'code' not in params:
            params['code'] = 'testSystemKey'
        headers = dict(kwargs.pop('headers', {}))
        headers['aeg-event-type'] = 'Notification'
        return request_method(url, *args, params=params, headers=headers,
                              **kwargs)

    @testutils.retryable_test(3, 5)
    @unittest.skip("Run locally. Running on Azure fails with 401/403 as the"
                   "host does not pick up the SecretKey from the"
                   "azure_functions_worker.testutils.py.SECRETS_TEMPLATE and"
                   "because of which we cannot test eventGrid webhook"
                   "invocation correctly.")
    def test_eventgrid_trigger(self):
        """test event_grid trigger

        This test calls the eventgrid_trigger function, sends in `data` as body
        to the webhook for eventgrid. Once the event is received, the function
        writes the data to the blob store.

        Then get_eventgrid_triggered gets called (httpTrigger) and takes blob
        input binding, reading the previously written text in blob store
        `python-worker-tests/test-eventgrid-triggered.txt`, and then we validate
        that the written text matches the one passed to the eventgrid trigger.
        """
        data = [{
            "topic": "test-topic",
            "subject": "test-subject",
            "eventType": "Microsoft.Storage.BlobCreated",
            "eventTime": "2018-01-01T00:00:00.000000123Z",
            "id": str(uuid.uuid4()),
            "data": {
                "api": "PutBlockList",
                "clientRequestId": "2c169f2f-7b3b-4d99-839b-c92a2d25801b",
                "requestId": "44d4f022-001e-003c-466b-940cba000000",
                "eTag": "0x8D562831044DDD0",
                "contentType": "application/octet-stream",
                "contentLength": 2248,
                "blobType": "BlockBlob",
                "ur1": "foo",
                "sequencer": "000000000000272D000000000003D60F",
                "storageDiagnostics": {
                    "batchId": "b4229b3a-4d50-4ff4-a9f2-039ccf26efe9"
                }
            },
            "dataVersion": "",
            "metadataVersion": "1"
        }]

        r = self.eventgrid_webhook_request('POST', 'eventgrid_trigger',
                                           json=data)
        self.assertEqual(r.status_code, 202)

        max_retries = 10

        for try_no in range(max_retries):
            # Allow trigger to fire.
            time.sleep(2)

            try:
                # Check that the trigger has fired.
                r = self.webhost.request('GET', 'get_eventgrid_triggered')
                self.assertEqual(r.status_code, 200)

                response = r.json()
                self.assertLessEqual(response.items(), data[0].items())
            except AssertionError:
                if try_no == max_retries - 1:
                    raise
            else:
                break

    @testutils.retryable_test(1, 5)
    def test_eventgrid_output_binding(self):
        """test event_grid output binding

        This test needs three functions to work.
        1. `eventgrid_output_binding`
        2. `eventgrid_output_binding_message_to_blobstore`
        3. `eventgrid_output_binding_success`

        This test calls the eventgrid_output_binding function, sends in a unique
        uuid as `data` in the body to the httpTrigger which sends in that value
        in the eventGrid output data. The eventGrid topic is configured to
        send the event to a storage queue.

        The second function (`eventgrid_output_binding_message_to_blobstore`)
        reads from that storage queue and puts into a blob store.

        The third function (`eventgrid_output_binding_success`) reads the
        text from the blob store and compares with the expected result. The
        unique uuid should confirm if the message went through correctly to
        EventGrid and came back as a blob.
        """

        test_uuid = uuid.uuid4().__str__()
        expected_response = "Sent event with subject: {}, id: {}, data: {}, " \
                            "event_type: {} to EventGrid!".format(
                                "test-subject", "test-id",
                                f"{{'test_uuid': '{test_uuid}'}}",
                                "test-event-1")
        expected_final_data = {
            'id': 'test-id', 'subject': 'test-subject', 'dataVersion': '1.0',
            'eventType': 'test-event-1',
            'data': {'test_uuid': test_uuid}
        }

        r = self.webhost.request('GET', 'eventgrid_output_binding',
                                 params={'test_uuid': test_uuid})
        self.assertEqual(r.status_code, 200)
        response = r.text

        self.assertEqual(expected_response, response)

        max_retries = 10
        for try_no in range(max_retries):
            # Allow trigger to fire.
            time.sleep(2)

            try:
                # Check that the trigger has fired.
                r = self.webhost.request('GET',
                                         'eventgrid_output_binding_success')
                self.assertEqual(r.status_code, 200)
                response = r.json()

                # list of fields to check are limited as other fields contain
                # datetime or other uncertain values
                for f in ['data', 'id', 'eventType', 'subject', 'dataVersion']:
                    self.assertEqual(response[f], expected_final_data[f])

            except AssertionError:
                if try_no == max_retries - 1:
                    raise
            else:
                break



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import typing

from azure_functions_worker import testutils
from azure_functions_worker.testutils import TESTS_ROOT, remove_path

HOST_JSON_TEMPLATE_WITH_LOGLEVEL_INFO = """\
{
    "version": "2.0",
    "logging": {
        "logLevel": {
           "default": "Information"
        }
    },
    "functionTimeout": "00:05:00"
}
"""


class TestLogFilteringFunctions(testutils.WebHostTestCase):

    @classmethod
    def setUpClass(cls):
        host_json = TESTS_ROOT / cls.get_script_dir() / 'host.json'

        with open(host_json, 'w+') as f:
            f.write(HOST_JSON_TEMPLATE_WITH_LOGLEVEL_INFO)

        super(TestLogFilteringFunctions, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        host_json = TESTS_ROOT / cls.get_script_dir() / 'host.json'
        remove_path(host_json)

        super(TestLogFilteringFunctions, cls).tearDownClass()

    @classmethod
    def get_script_dir(cls):
        return testutils.UNIT_TESTS_FOLDER / 'log_filtering_functions'

    def test_debug_logging(self):
        r = self.webhost.request('GET', 'debug_logging')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.text, 'OK-debug')

    def check_log_debug_logging(self, host_out: typing.List[str]):
        self.assertIn('logging info', host_out)
        self.assertIn('logging warning', host_out)
        self.assertIn('logging error', host_out)
        self.assertNotIn('logging debug', host_out)

    def test_debug_with_user_logging(self):
        r = self.webhost.request('GET', 'debug_user_logging')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.text, 'OK-user-debug')

    def check_log_debug_with_user_logging(self, host_out: typing.List[str]):
        self.assertIn('logging info', host_out)
        self.assertIn('logging warning', host_out)
        self.assertIn('logging error', host_out)
        self.assertNotIn('logging debug', host_out)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from azure_functions_worker import protos
from azure_functions_worker import testutils


class TestDurableFunctions(testutils.AsyncTestCase):
    durable_functions_dir = testutils.UNIT_TESTS_FOLDER / 'durable_functions'

    async def test_mock_activity_trigger(self):
        async with testutils.start_mockhost(
                script_root=self.durable_functions_dir) as host:

            func_id, r = await host.load_function('activity_trigger')

            self.assertEqual(r.response.function_id, func_id)
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)

            _, r = await host.invoke_function(
                'activity_trigger', [
                    # According to Durable Python
                    # Activity Trigger's input must be json serializable
                    protos.ParameterBinding(
                        name='input',
                        data=protos.TypedData(
                            string='test single_word'
                        )
                    )
                ]
            )
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)
            self.assertEqual(
                r.response.return_value,
                protos.TypedData(json='"test single_word"')
            )

    async def test_mock_activity_trigger_no_anno(self):
        async with testutils.start_mockhost(
                script_root=self.durable_functions_dir) as host:

            func_id, r = await host.load_function('activity_trigger_no_anno')

            self.assertEqual(r.response.function_id, func_id)
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)

            _, r = await host.invoke_function(
                'activity_trigger_no_anno', [
                    # According to Durable Python
                    # Activity Trigger's input must be json serializable
                    protos.ParameterBinding(
                        name='input',
                        data=protos.TypedData(
                            string='test multiple words'
                        )
                    )
                ]
            )
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)
            self.assertEqual(
                r.response.return_value,
                protos.TypedData(json='"test multiple words"')
            )

    async def test_mock_activity_trigger_dict(self):
        async with testutils.start_mockhost(
                script_root=self.durable_functions_dir) as host:

            func_id, r = await host.load_function('activity_trigger_dict')

            self.assertEqual(r.response.function_id, func_id)
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)

            _, r = await host.invoke_function(
                'activity_trigger_dict', [
                    # According to Durable Python
                    # Activity Trigger's input must be json serializable
                    protos.ParameterBinding(
                        name='input',
                        data=protos.TypedData(
                            json='{"bird": "Crane"}'
                        )
                    )
                ]
            )
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)
            self.assertEqual(
                r.response.return_value,
                protos.TypedData(json='{"bird": "enarC"}')
            )

    async def test_mock_activity_trigger_int_to_float(self):
        async with testutils.start_mockhost(
                script_root=self.durable_functions_dir) as host:

            func_id, r = await host.load_function(
                'activity_trigger_int_to_float')

            self.assertEqual(r.response.function_id, func_id)
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)

            _, r = await host.invoke_function(
                'activity_trigger_int_to_float', [
                    # According to Durable Python
                    # Activity Trigger's input must be json serializable
                    protos.ParameterBinding(
                        name='input',
                        data=protos.TypedData(
                            json=str(int(10))
                        )
                    )
                ]
            )
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)
            self.assertEqual(
                r.response.return_value,
                protos.TypedData(json='-11.0')
            )

    async def test_mock_orchestration_trigger(self):
        async with testutils.start_mockhost(
                script_root=self.durable_functions_dir) as host:

            func_id, r = await host.load_function('orchestration_trigger')

            self.assertEqual(r.response.function_id, func_id)
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)

            _, r = await host.invoke_function(
                'orchestration_trigger', [
                    protos.ParameterBinding(
                        name='context',
                        data=protos.TypedData(
                            string='Durable functions coming soon'
                        )
                    )
                ]
            )
            self.assertEqual(r.response.result.status,
                             protos.StatusResult.Success)
            self.assertEqual(
                r.response.return_value,
                protos.TypedData(json='Durable functions coming soon :)')
            )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import io

import azure.functions as azf


def main(req: azf.HttpRequest, file: azf.Out[io.StringIO]) -> str:
    file.set(io.StringIO('filelike'))
    return 'OK'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
"""GRPC client.

Implements loading and execution of Python workers.
"""

import asyncio
import concurrent.futures
import importlib
import inspect
import logging
import os
import queue
import sys
import threading
from asyncio import BaseEventLoop
from logging import LogRecord
from typing import Optional, List

import grpc

from . import bindings
from . import constants
from . import functions
from . import loader
from . import protos
from .constants import (CONSOLE_LOG_PREFIX, PYTHON_THREADPOOL_THREAD_COUNT,
                        PYTHON_THREADPOOL_THREAD_COUNT_DEFAULT,
                        PYTHON_THREADPOOL_THREAD_COUNT_MAX,
                        PYTHON_THREADPOOL_THREAD_COUNT_MIN)
from .logging import disable_console_logging, enable_console_logging
from .logging import error_logger, is_system_log_category, logger
from .utils.common import get_app_setting
from .utils.tracing import marshall_exception_trace
from .utils.wrappers import disable_feature_by

_TRUE = "true"

"""In Python 3.6, the current_task method was in the Task class, but got moved
out in 3.7+ and fully removed in 3.9. Thus, to support 3.6 and 3.9 together, we
need to switch the implementation of current_task for 3.6.
"""
_CURRENT_TASK = asyncio.Task.current_task \
    if (sys.version_info[0] == 3 and sys.version_info[1] == 6) \
    else asyncio.current_task


class DispatcherMeta(type):

    __current_dispatcher__ = None

    @property
    def current(mcls):
        disp = mcls.__current_dispatcher__
        if disp is None:
            raise RuntimeError('no currently running Dispatcher is found')
        return disp


class Dispatcher(metaclass=DispatcherMeta):

    _GRPC_STOP_RESPONSE = object()

    def __init__(self, loop: BaseEventLoop, host: str, port: int,
                 worker_id: str, request_id: str,
                 grpc_connect_timeout: float,
                 grpc_max_msg_len: int = -1) -> None:
        self._loop = loop
        self._host = host
        self._port = port
        self._request_id = request_id
        self._worker_id = worker_id
        self._functions = functions.Registry()

        self._old_task_factory = None

        # We allow the customer to change synchronous thread pool count by
        # PYTHON_THREADPOOL_THREAD_COUNT app setting. The default value is 1.
        self._sync_tp_max_workers: int = self._get_sync_tp_max_workers()
        self._sync_call_tp: concurrent.futures.Executor = (
            self._create_sync_call_tp(self._sync_tp_max_workers)
        )

        self._grpc_connect_timeout: float = grpc_connect_timeout
        # This is set to -1 by default to remove the limitation on msg size
        self._grpc_max_msg_len: int = grpc_max_msg_len
        self._grpc_resp_queue: queue.Queue = queue.Queue()
        self._grpc_connected_fut = loop.create_future()
        self._grpc_thread: threading.Thread = threading.Thread(
            name='grpc-thread', target=self.__poll_grpc)

    @classmethod
    async def connect(cls, host: str, port: int, worker_id: str,
                      request_id: str, connect_timeout: float):
        loop = asyncio.events.get_event_loop()
        disp = cls(loop, host, port, worker_id, request_id, connect_timeout)
        disp._grpc_thread.start()
        await disp._grpc_connected_fut
        logger.info('Successfully opened gRPC channel to %s:%s ', host, port)
        return disp

    async def dispatch_forever(self):
        if DispatcherMeta.__current_dispatcher__ is not None:
            raise RuntimeError('there can be only one running dispatcher per '
                               'process')

        self._old_task_factory = self._loop.get_task_factory()

        loader.install()

        DispatcherMeta.__current_dispatcher__ = self
        try:
            forever = self._loop.create_future()

            self._grpc_resp_queue.put_nowait(
                protos.StreamingMessage(
                    request_id=self.request_id,
                    start_stream=protos.StartStream(
                        worker_id=self.worker_id)))

            self._loop.set_task_factory(
                lambda loop, coro: ContextEnabledTask(coro, loop=loop))

            # Detach console logging before enabling GRPC channel logging
            logger.info('Detaching console logging.')
            disable_console_logging()

            # Attach gRPC logging to the root logger. Since gRPC channel is
            # established, should use it for system and user logs
            logging_handler = AsyncLoggingHandler()
            root_logger = logging.getLogger()

            # Don't change this unless you read #780 and #745
            root_logger.setLevel(logging.INFO)
            root_logger.addHandler(logging_handler)
            logger.info('Switched to gRPC logging.')
            logging_handler.flush()

            try:
                await forever
            finally:
                logger.warning('Detaching gRPC logging due to exception.')
                logging_handler.flush()
                root_logger.removeHandler(logging_handler)

                # Reenable console logging when there's an exception
                enable_console_logging()
                logger.warning('Switched to console logging due to exception.')
        finally:
            DispatcherMeta.__current_dispatcher__ = None

            loader.uninstall()

            self._loop.set_task_factory(self._old_task_factory)
            self.stop()

    def stop(self) -> None:
        if self._grpc_thread is not None:
            self._grpc_resp_queue.put_nowait(self._GRPC_STOP_RESPONSE)
            self._grpc_thread.join()
            self._grpc_thread = None

        self._stop_sync_call_tp()

    def on_logging(self, record: logging.LogRecord, formatted_msg: str) -> None:
        if record.levelno >= logging.CRITICAL:
            log_level = protos.RpcLog.Critical
        elif record.levelno >= logging.ERROR:
            log_level = protos.RpcLog.Error
        elif record.levelno >= logging.WARNING:
            log_level = protos.RpcLog.Warning
        elif record.levelno >= logging.INFO:
            log_level = protos.RpcLog.Information
        elif record.levelno >= logging.DEBUG:
            log_level = protos.RpcLog.Debug
        else:
            log_level = getattr(protos.RpcLog, 'None')

        if is_system_log_category(record.name):
            log_category = protos.RpcLog.RpcLogCategory.Value('System')
        else:  # customers using logging will yield 'root' in record.name
            log_category = protos.RpcLog.RpcLogCategory.Value('User')

        log = dict(
            level=log_level,
            message=formatted_msg,
            category=record.name,
            log_category=log_category
        )

        invocation_id = get_current_invocation_id()
        if invocation_id is not None:
            log['invocation_id'] = invocation_id

        # XXX: When an exception field is set in RpcLog, WebHost doesn't
        # wait for the call result and simply aborts the execution.
        #
        # if record.exc_info and record.exc_info[1] is not None:
        #     log['exception'] = self._serialize_exception(record.exc_info[1])

        self._grpc_resp_queue.put_nowait(
            protos.StreamingMessage(
                request_id=self.request_id,
                rpc_log=protos.RpcLog(**log)))

    @property
    def request_id(self) -> str:
        return self._request_id

    @property
    def worker_id(self) -> str:
        return self._worker_id

    # noinspection PyBroadException
    @staticmethod
    def _serialize_exception(exc: Exception):
        try:
            message = f'{type(exc).__name__}: {exc}'
        except Exception:
            message = ('Unhandled exception in function. '
                       'Could not serialize original exception message.')

        try:
            stack_trace = marshall_exception_trace(exc)
        except Exception:
            stack_trace = ''

        return protos.RpcException(message=message, stack_trace=stack_trace)

    async def _dispatch_grpc_request(self, request):
        content_type = request.WhichOneof('content')
        request_handler = getattr(self, f'_handle__{content_type}', None)
        if request_handler is None:
            # Don't crash on unknown messages.  Some of them can be ignored;
            # and if something goes really wrong the host can always just
            # kill the worker's process.
            logger.error(f'unknown StreamingMessage content type '
                         f'{content_type}')
            return

        resp = await request_handler(request)
        self._grpc_resp_queue.put_nowait(resp)

    async def _handle__worker_init_request(self, req):
        logger.info('Received WorkerInitRequest, request ID %s',
                    self.request_id)

        capabilities = {
            constants.RAW_HTTP_BODY_BYTES: _TRUE,
            constants.TYPED_DATA_COLLECTION: _TRUE,
            constants.RPC_HTTP_BODY_ONLY: _TRUE,
            constants.RPC_HTTP_TRIGGER_METADATA_REMOVED: _TRUE,
        }

        return protos.StreamingMessage(
            request_id=self.request_id,
            worker_init_response=protos.WorkerInitResponse(
                capabilities=capabilities,
                result=protos.StatusResult(
                    status=protos.StatusResult.Success)))

    async def _handle__function_load_request(self, req):
        func_request = req.function_load_request
        function_id = func_request.function_id

        logger.info(f'Received FunctionLoadRequest, '
                    f'request ID: {self.request_id}, '
                    f'function ID: {function_id}')
        try:
            func = loader.load_function(
                func_request.metadata.name,
                func_request.metadata.directory,
                func_request.metadata.script_file,
                func_request.metadata.entry_point)

            self._functions.add_function(
                function_id, func, func_request.metadata)

            logger.info('Successfully processed FunctionLoadRequest, '
                        f'request ID: {self.request_id}, '
                        f'function ID: {function_id}')

            return protos.StreamingMessage(
                request_id=self.request_id,
                function_load_response=protos.FunctionLoadResponse(
                    function_id=function_id,
                    result=protos.StatusResult(
                        status=protos.StatusResult.Success)))

        except Exception as ex:
            return protos.StreamingMessage(
                request_id=self.request_id,
                function_load_response=protos.FunctionLoadResponse(
                    function_id=function_id,
                    result=protos.StatusResult(
                        status=protos.StatusResult.Failure,
                        exception=self._serialize_exception(ex))))

    async def _handle__invocation_request(self, req):
        invoc_request = req.invocation_request

        invocation_id = invoc_request.invocation_id
        function_id = invoc_request.function_id
        trace_context = bindings.TraceContext(
            invoc_request.trace_context.trace_parent,
            invoc_request.trace_context.trace_state,
            invoc_request.trace_context.attributes)
        # Set the current `invocation_id` to the current task so
        # that our logging handler can find it.
        current_task = _CURRENT_TASK(self._loop)
        assert isinstance(current_task, ContextEnabledTask)
        current_task.set_azure_invocation_id(invocation_id)

        try:
            fi: functions.FunctionInfo = self._functions.get_function(
                function_id)

            function_invocation_logs: List[str] = [
                'Received FunctionInvocationRequest',
                f'request ID: {self.request_id}',
                f'function ID: {function_id}',
                f'invocation ID: {invocation_id}',
                f'function type: {"async" if fi.is_async else "sync"}'
            ]
            if not fi.is_async:
                function_invocation_logs.append(
                    f'sync threadpool max workers: {self._sync_tp_max_workers}'
                )
            logger.info(', '.join(function_invocation_logs))

            args = {}
            for pb in invoc_request.input_data:
                pb_type_info = fi.input_types[pb.name]
                if bindings.is_trigger_binding(pb_type_info.binding_name):
                    trigger_metadata = invoc_request.trigger_metadata
                else:
                    trigger_metadata = None
                args[pb.name] = bindings.from_incoming_proto(
                    pb_type_info.binding_name, pb.data,
                    trigger_metadata=trigger_metadata,
                    pytype=pb_type_info.pytype)

            if fi.requires_context:
                args['context'] = bindings.Context(
                    fi.name, fi.directory, invocation_id, trace_context)

            if fi.output_types:
                for name in fi.output_types:
                    args[name] = bindings.Out()

            if fi.is_async:
                call_result = await fi.func(**args)
            else:
                call_result = await self._loop.run_in_executor(
                    self._sync_call_tp,
                    self.__run_sync_func, invocation_id, fi.func, args)
            if call_result is not None and not fi.has_return:
                raise RuntimeError(f'function {fi.name!r} without a $return '
                                   'binding returned a non-None value')

            output_data = []
            if fi.output_types:
                for out_name, out_type_info in fi.output_types.items():
                    val = args[out_name].get()
                    if val is None:
                        # TODO: is the "Out" parameter optional?
                        # Can "None" be marshaled into protos.TypedData?
                        continue

                    rpc_val = bindings.to_outgoing_proto(
                        out_type_info.binding_name, val,
                        pytype=out_type_info.pytype)
                    assert rpc_val is not None

                    output_data.append(
                        protos.ParameterBinding(
                            name=out_name,
                            data=rpc_val))

            return_value = None
            if fi.return_type is not None:
                return_value = bindings.to_outgoing_proto(
                    fi.return_type.binding_name, call_result,
                    pytype=fi.return_type.pytype)

            # Actively flush customer print() function to console
            sys.stdout.flush()

            return protos.StreamingMessage(
                request_id=self.request_id,
                invocation_response=protos.InvocationResponse(
                    invocation_id=invocation_id,
                    return_value=return_value,
                    result=protos.StatusResult(
                        status=protos.StatusResult.Success),
                    output_data=output_data))

        except Exception as ex:
            return protos.StreamingMessage(
                request_id=self.request_id,
                invocation_response=protos.InvocationResponse(
                    invocation_id=invocation_id,
                    result=protos.StatusResult(
                        status=protos.StatusResult.Failure,
                        exception=self._serialize_exception(ex))))

    async def _handle__function_environment_reload_request(self, req):
        """Only runs on Linux Consumption placeholder specialization.
        """
        try:
            logger.info('Received FunctionEnvironmentReloadRequest, '
                        'request ID: %s', self.request_id)

            func_env_reload_request = req.function_environment_reload_request

            # Import before clearing path cache so that the default
            # azure.functions modules is available in sys.modules for
            # customer use
            import azure.functions # NoQA

            # Append function project root to module finding sys.path
            if func_env_reload_request.function_app_directory:
                sys.path.append(func_env_reload_request.function_app_directory)

            # Clear sys.path import cache, reload all module from new sys.path
            sys.path_importer_cache.clear()

            # Reload environment variables
            os.environ.clear()
            env_vars = func_env_reload_request.environment_variables
            for var in env_vars:
                os.environ[var] = env_vars[var]

            # Apply PYTHON_THREADPOOL_THREAD_COUNT
            self._stop_sync_call_tp()
            self._sync_tp_max_workers = self._get_sync_tp_max_workers()
            self._sync_call_tp = (
                self._create_sync_call_tp(self._sync_tp_max_workers)
            )

            # Reload package namespaces for customer's libraries
            packages_to_reload = ['azure', 'google']
            for p in packages_to_reload:
                try:
                    logger.info(f'Reloading {p} module')
                    importlib.reload(sys.modules[p])
                except Exception as ex:
                    logger.info('Unable to reload {}: \n{}'.format(p, ex))
                logger.info(f'Reloaded {p} module')

            # Reload azure.functions to give user package precedence
            logger.info('Reloading azure.functions module at %s',
                        inspect.getfile(sys.modules['azure.functions']))
            try:
                importlib.reload(sys.modules['azure.functions'])
                logger.info('Reloaded azure.functions module now at %s',
                            inspect.getfile(sys.modules['azure.functions']))
            except Exception as ex:
                logger.info('Unable to reload azure.functions. '
                            'Using default. Exception:\n{}'.format(ex))

            # Change function app directory
            if getattr(func_env_reload_request,
                       'function_app_directory', None):
                self._change_cwd(
                    func_env_reload_request.function_app_directory)

            success_response = protos.FunctionEnvironmentReloadResponse(
                result=protos.StatusResult(
                    status=protos.StatusResult.Success))

            return protos.StreamingMessage(
                request_id=self.request_id,
                function_environment_reload_response=success_response)

        except Exception as ex:
            failure_response = protos.FunctionEnvironmentReloadResponse(
                result=protos.StatusResult(
                    status=protos.StatusResult.Failure,
                    exception=self._serialize_exception(ex)))

            return protos.StreamingMessage(
                request_id=self.request_id,
                function_environment_reload_response=failure_response)

    @disable_feature_by(constants.PYTHON_ROLLBACK_CWD_PATH)
    def _change_cwd(self, new_cwd: str):
        if os.path.exists(new_cwd):
            os.chdir(new_cwd)
            logger.info('Changing current working directory to %s', new_cwd)
        else:
            logger.warning('Directory %s is not found when reloading', new_cwd)

    def _stop_sync_call_tp(self):
        """Deallocate the current synchronous thread pool and assign
        self._sync_call_tp to None. If the thread pool does not exist,
        this will be a no op.
        """
        if getattr(self, '_sync_call_tp', None):
            self._sync_call_tp.shutdown()
            self._sync_call_tp = None

    def _get_sync_tp_max_workers(self) -> int:
        def tp_max_workers_validator(value: str) -> bool:
            try:
                int_value = int(value)
            except ValueError:
                logger.warning(f'{PYTHON_THREADPOOL_THREAD_COUNT} must be an '
                               'integer')
                return False

            if int_value < PYTHON_THREADPOOL_THREAD_COUNT_MIN or (
               int_value > PYTHON_THREADPOOL_THREAD_COUNT_MAX):
                logger.warning(f'{PYTHON_THREADPOOL_THREAD_COUNT} must be set '
                               'to a value between 1 and 32')
                return False

            return True

        return int(get_app_setting(
            setting=PYTHON_THREADPOOL_THREAD_COUNT,
            default_value=f'{PYTHON_THREADPOOL_THREAD_COUNT_DEFAULT}',
            validator=tp_max_workers_validator))

    def _create_sync_call_tp(
            self, max_worker: int) -> concurrent.futures.Executor:
        """Create a thread pool executor with max_worker. This is a wrapper
        over ThreadPoolExecutor constructor. Consider calling this method after
        _stop_sync_call_tp() to ensure only 1 synchronous thread pool is
        running.
        """
        return concurrent.futures.ThreadPoolExecutor(
            max_workers=max_worker
        )

    def __run_sync_func(self, invocation_id, func, params):
        # This helper exists because we need to access the current
        # invocation_id from ThreadPoolExecutor's threads.
        _invocation_id_local.v = invocation_id
        try:
            return func(**params)
        finally:
            _invocation_id_local.v = None

    def __poll_grpc(self):
        options = []
        if self._grpc_max_msg_len:
            options.append(('grpc.max_receive_message_length',
                            self._grpc_max_msg_len))
            options.append(('grpc.max_send_message_length',
                            self._grpc_max_msg_len))

        channel = grpc.insecure_channel(
            f'{self._host}:{self._port}', options)

        try:
            grpc.channel_ready_future(channel).result(
                timeout=self._grpc_connect_timeout)
        except Exception as ex:
            self._loop.call_soon_threadsafe(
                self._grpc_connected_fut.set_exception, ex)
            return
        else:
            self._loop.call_soon_threadsafe(
                self._grpc_connected_fut.set_result, True)

        stub = protos.FunctionRpcStub(channel)

        def gen(resp_queue):
            while True:
                msg = resp_queue.get()
                if msg is self._GRPC_STOP_RESPONSE:
                    grpc_req_stream.cancel()
                    return
                yield msg

        grpc_req_stream = stub.EventStream(gen(self._grpc_resp_queue))
        try:
            for req in grpc_req_stream:
                self._loop.call_soon_threadsafe(
                    self._loop.create_task, self._dispatch_grpc_request(req))
        except Exception as ex:
            if ex is grpc_req_stream:
                # Yes, this is how grpc_req_stream iterator exits.
                return
            error_logger.exception('unhandled error in gRPC thread')
            raise


class AsyncLoggingHandler(logging.Handler):

    def emit(self, record: LogRecord) -> None:
        # Since we disable console log after gRPC channel is initiated,
        # we should redirect all the messages into dispatcher.

        # When dispatcher receives an exception, it should switch back
        # to console logging. However, it is possible that
        # __current_dispatcher__ is set to None as there are still messages
        # buffered in this handler, not calling the emit yet.
        msg = self.format(record)
        try:
            Dispatcher.current.on_logging(record, msg)
        except RuntimeError as runtime_error:
            # This will cause 'Dispatcher not found' failure.
            # Logging such of an issue will cause infinite loop of gRPC logging
            # To mitigate, we should suppress the 2nd level error logging here
            # and use print function to report exception instead.
            print(f'{CONSOLE_LOG_PREFIX} ERROR: {str(runtime_error)}',
                  file=sys.stderr, flush=True)


class ContextEnabledTask(asyncio.Task):

    AZURE_INVOCATION_ID = '__azure_function_invocation_id__'

    def __init__(self, coro, loop):
        super().__init__(coro, loop=loop)

        current_task = _CURRENT_TASK(loop)
        if current_task is not None:
            invocation_id = getattr(
                current_task, self.AZURE_INVOCATION_ID, None)
            if invocation_id is not None:
                self.set_azure_invocation_id(invocation_id)

    def set_azure_invocation_id(self, invocation_id: str) -> None:
        setattr(self, self.AZURE_INVOCATION_ID, invocation_id)


def get_current_invocation_id() -> Optional[str]:
    loop = asyncio._get_running_loop()
    if loop is not None:
        current_task = _CURRENT_TASK(loop)
        if current_task is not None:
            task_invocation_id = getattr(current_task,
                                         ContextEnabledTask.AZURE_INVOCATION_ID,
                                         None)
            if task_invocation_id is not None:
                return task_invocation_id

    return getattr(_invocation_id_local, 'v', None)


_invocation_id_local = threading.local()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as func


def main(req: func.HttpRequest, doc: func.Out[func.Document]):
    doc.set(func.Document.from_json(req.get_body()))

    return 'OK'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import json
import time
import pathlib
from datetime import datetime
from dateutil import parser, tz

from azure_functions_worker import testutils


class TestEventHubFunctions(testutils.WebHostTestCase):
    """Test EventHub Trigger and Output Bindings (cardinality: many).

    Each testcase consists of 3 part:
    1. An eventhub_output_batch HTTP trigger for generating EventHub event
    2. An eventhub_multiple EventHub trigger for converting event into blob
    3. A get_eventhub_batch_triggered HTTP trigger for the event body
    """

    @classmethod
    def get_script_dir(cls):
        return testutils.E2E_TESTS_FOLDER / 'eventhub_batch_functions'

    @testutils.retryable_test(3, 5)
    def test_eventhub_multiple(self):
        NUM_EVENTS = 3
        all_row_keys_seen = dict([(str(i), True) for i in range(NUM_EVENTS)])
        partition_key = str(round(time.time()))

        # Dynamically rewrite function.json to point to new partition key
        # for recording EventHub state
        old_partition_key = self._get_table_partition_key()
        self._set_table_partition_key(partition_key)

        # wait for host to restart after change
        time.sleep(5)

        docs = []
        for i in range(NUM_EVENTS):
            doc = {'PartitionKey': partition_key, 'RowKey': i}
            docs.append(doc)

        r = self.webhost.request('POST', 'eventhub_output_batch',
                                 data=json.dumps(docs))
        self.assertEqual(r.status_code, 200)

        row_keys = [str(i) for i in range(NUM_EVENTS)]
        seen = [False] * NUM_EVENTS
        row_keys_seen = dict(zip(row_keys, seen))

        # Allow trigger to fire.
        time.sleep(5)

        try:
            r = self.webhost.request('GET', 'get_eventhub_batch_triggered')
            self.assertEqual(r.status_code, 200)
            entries = r.json()
            for entry in entries:
                self.assertEqual(entry['PartitionKey'], partition_key)
                row_key = entry['RowKey']
                row_keys_seen[row_key] = True

            self.assertDictEqual(all_row_keys_seen, row_keys_seen)
        finally:
            self._cleanup(old_partition_key)

    @testutils.retryable_test(3, 5)
    def test_eventhub_multiple_with_metadata(self):
        # Generate a unique event body for EventHub event
        # Record the start_time and end_time for checking event enqueue time
        start_time = datetime.now(tz=tz.UTC)
        count = 10
        random_number = str(round(time.time()) % 1000)
        req_body = {
            'body': random_number
        }

        # Invoke metadata_output HttpTrigger to generate an EventHub event
        # from azure-eventhub SDK
        r = self.webhost.request('POST',
                                 f'metadata_output_batch?count={count}',
                                 data=json.dumps(req_body))
        self.assertEqual(r.status_code, 200)
        self.assertIn('OK', r.text)
        end_time = datetime.now(tz=tz.UTC)

        # Once the event get generated, allow function host to pool from
        # EventHub and wait for metadata_multiple to execute,
        # converting the event metadata into a blob.
        time.sleep(5)

        # Call get_metadata_batch_triggered to retrieve event metadata
        r = self.webhost.request('GET', 'get_metadata_batch_triggered')
        self.assertEqual(r.status_code, 200)

        # Check metadata and events length, events should be batch processed
        events = r.json()
        self.assertIsInstance(events, list)
        self.assertGreater(len(events), 1)

        # EventhubEvent property check
        for event_index in range(len(events)):
            event = events[event_index]

            # Check if the event is enqueued between start_time and end_time
            enqueued_time = parser.isoparse(event['enqueued_time'])
            self.assertTrue(start_time < enqueued_time < end_time)

            # Check if event properties are properly set
            self.assertIsNone(event['partition_key'])  # only 1 partition
            self.assertGreaterEqual(event['sequence_number'], 0)
            self.assertIsNotNone(event['offset'])

            # Check if event.metadata field is properly set
            self.assertIsNotNone(event['metadata'])
            metadata = event['metadata']
            sys_props_array = metadata['SystemPropertiesArray']
            sys_props = sys_props_array[event_index]
            enqueued_time = parser.isoparse(sys_props['EnqueuedTimeUtc'])

            # Check event trigger time and other system properties
            self.assertTrue(start_time < enqueued_time < end_time)
            self.assertIsNone(sys_props['PartitionKey'])
            self.assertGreaterEqual(sys_props['SequenceNumber'], 0)
            self.assertIsNotNone(sys_props['Offset'])

            # Metadata check eventhub trigger name
            self.assertEqual(metadata['sys']['MethodName'],
                             'metadata_multiple')

    def _cleanup(self, old_partition_key):
        self._set_table_partition_key(old_partition_key)

    def _get_table_partition_key(self):
        func_dict = self._get_table_function_json_dict()
        partition_key = func_dict['bindings'][1]['partitionKey']
        return partition_key

    def _set_table_partition_key(self, partition_key):
        full_json_path = self._get_table_function_json_path()

        func_dict = self._get_table_function_json_dict()
        func_dict['bindings'][1]['partitionKey'] = partition_key

        with open(full_json_path, 'w') as f:
            json.dump(func_dict, f, indent=2)

    def _get_table_function_json_dict(self):
        full_json_path = self._get_table_function_json_path()

        with open(full_json_path, 'r') as f:
            func_dict = json.load(f)

        return func_dict

    def _get_table_function_json_path(self):
        script_dir = pathlib.Path(self.get_script_dir())
        json_path = pathlib.Path('get_eventhub_batch_triggered/function.json')
        return testutils.TESTS_ROOT / script_dir / json_path



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import asyncio
import pathlib
import subprocess
import sys
import textwrap

from azure_functions_worker import testutils


class TestLoader(testutils.WebHostTestCase):

    @classmethod
    def get_script_dir(cls):
        return testutils.UNIT_TESTS_FOLDER / 'load_functions'

    def test_loader_simple(self):
        r = self.webhost.request('GET', 'simple')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.text, '__app__.simple.main')

    def test_loader_custom_entrypoint(self):
        r = self.webhost.request('GET', 'entrypoint')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.text, '__app__.entrypoint.main')

    def test_loader_subdir(self):
        r = self.webhost.request('GET', 'subdir')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.text, '__app__.subdir.sub.main')

    def test_loader_relimport(self):
        r = self.webhost.request('GET', 'relimport')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.text, '__app__.relimport.relative')

    def test_loader_submodule(self):
        r = self.webhost.request('GET', 'submodule')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.text, '__app__.submodule.sub_module.module')

    def test_loader_parentmodule(self):
        r = self.webhost.request('GET', 'parentmodule')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.text, '__app__.parentmodule.module')

    def test_loader_absolute_thirdparty(self):
        """Allow third-party package import from .python_packages
        and worker_venv
        """

        r = self.webhost.request('GET', 'absolute_thirdparty')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.text, 'eh = azure.eventhub')

    def test_loader_prioritize_customer_module(self):
        """When a module in customer code has the same name with a third-party
        package, the worker should prioritize third-party package
        """

        r = self.webhost.request('GET', 'name_collision')
        self.assertEqual(r.status_code, 200)
        self.assertRegex(r.text, r'pt.__version__ = \d+.\d+.\d+')

    def test_loader_fix_customer_module_with_app_import(self):
        """When a module in customer code has the same name with a third-party
        package, if customer uses "import __app__.<module>" statement,
        the worker should load customer package
        """

        r = self.webhost.request('GET', 'name_collision_app_import')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.text, 'pt.__version__ = from.customer.code')

    def test_loader_implicit_import(self):
        """Since sys.path is now fixed with script root appended,
        implicit import statement is now acceptable.
        """

        r = self.webhost.request('GET', 'implicit_import')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.text, 's_main = simple.main')

    def test_loader_module_not_found(self):
        """If a module cannot be found, should throw an exception with
        trouble shooting link https://aka.ms/functions-modulenotfound
        """
        r = self.webhost.request('GET', 'module_not_found')
        self.assertEqual(r.status_code, 500)

    def check_log_loader_module_not_found(self, host_out):
        self.assertIn("Exception: ModuleNotFoundError: "
                      "No module named 'notfound'. "
                      "Troubleshooting Guide: "
                      "https://aka.ms/functions-modulenotfound", host_out)


class TestPluginLoader(testutils.AsyncTestCase):

    async def test_entry_point_plugin(self):
        test_binding = pathlib.Path(__file__).parent / 'test-binding'
        subprocess.run([
            sys.executable, '-m', 'pip',
            '--disable-pip-version-check',
            'install', '--quiet',
            '-e', test_binding
        ], check=True)

        # This test must be run in a subprocess so that
        # pkg_resources picks up the newly installed package.
        code = textwrap.dedent('''
import asyncio
from azure_functions_worker import protos
from azure_functions_worker import testutils

async def _runner():
    async with testutils.start_mockhost(
            script_root='unittests/test-binding/functions') as host:
        func_id, r = await host.load_function('foo')

        print(r.response.function_id == func_id)
        print(r.response.result.status == protos.StatusResult.Success)

asyncio.get_event_loop().run_until_complete(_runner())
''')

        try:
            proc = await asyncio.create_subprocess_exec(
                sys.executable, '-c', code,
                stdout=asyncio.subprocess.PIPE)

            stdout, stderr = await proc.communicate()

            # Trimming off carriage return charater when testing on Windows
            stdout_lines = [
                line.replace(b'\r', b'') for line in stdout.strip().split(b'\n')
            ]
            self.assertEqual(stdout_lines, [b'True', b'True'])

        finally:
            subprocess.run([
                sys.executable, '-m', 'pip',
                '--disable-pip-version-check',
                'uninstall', '-y', '--quiet', 'foo-binding'
            ], check=True)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as func


# This is testing the function load feature for the single event annotation
def main(event: func.EventHubEvent) -> str:
    return 'OK_ONE'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from .common import is_envvar_true
from .tracing import extend_exception_message
from typing import Callable, Optional


def enable_feature_by(flag: str, default: Optional[int] = None) -> Callable:
    def decorate(func):
        def call(*args, **kwargs):
            if is_envvar_true(flag):
                return func(*args, **kwargs)
            return default
        return call
    return decorate


def disable_feature_by(flag: str, default: None = None) -> Callable:
    def decorate(func):
        def call(*args, **kwargs):
            if not is_envvar_true(flag):
                return func(*args, **kwargs)
            return default
        return call
    return decorate


def attach_message_to_exception(expt_type: Exception, message: str) -> Callable:
    def decorate(func):
        def call(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except expt_type as e:
                raise extend_exception_message(e, message)
        return call
    return decorate



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as azf


def main(req: azf.HttpResponse):  # should be azf.HttpRequest
    return 'trust me, it is OK!'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from typing import List


# This is testing the function load feature for the multiple events annotation
# The event shouldn't be List[str]
def main(events: List[str]) -> str:
    return 'BAD'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as azf


def main(req, ret: azf.Out[azf.HttpRequest]):
    return 'trust me, it is OK!'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as func


def main(req: func.HttpRequest, docs: func.DocumentList) -> str:
    return func.HttpResponse(docs[0].to_json(), mimetype='application/json')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import json


# This is an actual EventHub trigger which handles Eventhub events in batches.
# It serializes multiple event data into a json and store it into a blob.
def main(events):
    table_entries = []
    for event in events:
        json_entry = event.get_body()
        table_entry = json.loads(json_entry)
        table_entries.append(table_entry)

    table_json = json.dumps(table_entries)

    return table_json



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import os
import json

import azure.functions as func
from azure.eventhub import EventHubProducerClient, EventData


# An HttpTrigger to generating EventHub event from azure-eventhub SDK.
# Events generated from azure-eventhub contain the full metadata.
def main(req: func.HttpRequest):
    # Get event count from http request query parameter
    count = int(req.params.get('count', '1'))

    # Parse event metadata from http request
    json_string = req.get_body().decode('utf-8')
    event_dict = json.loads(json_string)

    # Create an EventHub Client and event batch
    client = EventHubProducerClient.from_connection_string(
        os.getenv('AzureWebJobsEventHubConnectionString'),
        eventhub_name='python-worker-ci-eventhub-batch-metadata')

    # Generate new event based on http request with full metadata
    event_data_batch = client.create_batch()
    random_number = int(event_dict.get('body', '0'))
    for i in range(count):
        event_data_batch.add(EventData(str(random_number + i)))

    # Send out event into event hub
    with client:
        client.send_batch(event_data_batch)

    return f'OK'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

# Import simple module with implicit statement should now be acceptable
# since sys.path is now appended with function script root
from simple.main import main as s_main


def implicitinmport(req) -> str:
    return f's_main = {s_main(req)}'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions


def main(req: azure.functions.HttpRequest, context) -> str:
    return 'Hello World!'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import typing
import json
import azure.functions as func


# This is an actual EventHub trigger. It picks a few of EventHub properties
# and converts them into a storage blob
async def main(event: func.EventHubEvent) -> bytes:
    event_dict: typing.Mapping[str, typing.Any] = {
        'body': event.get_body().decode('utf-8'),
        # Uncomment this when the EnqueuedTimeUtc is fixed in azure-functions
        # 'enqueued_time': event.enqueued_time.isoformat(),
        'partition_key': event.partition_key,
        'sequence_number': event.sequence_number,
        'offset': event.offset,
        'metadata': event.metadata
    }

    return json.dumps(event_dict)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as azf


def main(req: azf.HttpRequest, msg: azf.Out[str]):
    msg.set(req.get_body())

    return 'OK'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from typing import Tuple
import os
from unittest.mock import patch
from azure_functions_worker import protos
from azure_functions_worker import testutils
from azure_functions_worker.constants import PYTHON_THREADPOOL_THREAD_COUNT


class TestDispatcher(testutils.AsyncTestCase):
    dispatcher_funcs_dir = testutils.UNIT_TESTS_FOLDER / 'dispatcher_functions'

    def setUp(self):
        self._pre_env = dict(os.environ)

    def tearDown(self):
        os.environ.clear()
        os.environ.update(self._pre_env)

    async def test_dispatcher_sync_threadpool_default_worker(self):
        """Test if the sync threadpool has maximum worker count set to 1
        by default
        """
        ctrl = testutils.start_mockhost(script_root=self.dispatcher_funcs_dir)

        async with ctrl as host:
            await self._check_if_function_is_ok(host)
            self.assertEqual(ctrl._worker._sync_tp_max_workers, 1)
            self.assertIsNotNone(ctrl._worker._sync_call_tp)

            # Check if the dispatcher still function
            await self._check_if_function_is_ok(host)

    async def test_dispatcher_sync_threadpool_set_worker(self):
        """Test if the sync threadpool maximum worker can be set
        """
        # Configure thread pool max worker
        os.environ.update({PYTHON_THREADPOOL_THREAD_COUNT: '5'})
        ctrl = testutils.start_mockhost(script_root=self.dispatcher_funcs_dir)

        async with ctrl as host:
            await self._check_if_function_is_ok(host)
            self.assertEqual(ctrl._worker._sync_tp_max_workers, 5)
            self.assertIsNotNone(ctrl._worker._sync_call_tp)

            # Check if the dispatcher still function
            await self._check_if_function_is_ok(host)

    async def test_dispatcher_sync_threadpool_invalid_worker_count(self):
        """Test when sync threadpool maximum worker is set to an invalid value,
        the host should fallback to default value 1
        """
        # The @patch decorator does not work as expected and will suppress
        # any assertion failures in the async test cases.
        # Thus we're moving the patch() method to use the with syntax

        with patch('azure_functions_worker.dispatcher.logger') as mock_logger:
            # Configure thread pool max worker to an invalid value
            os.environ.update({PYTHON_THREADPOOL_THREAD_COUNT: 'invalid'})
            ctrl = testutils.start_mockhost(
                script_root=self.dispatcher_funcs_dir)

            async with ctrl as host:
                await self._check_if_function_is_ok(host)

                # Ensure the dispatcher sync threadpool should fallback to 1
                self.assertEqual(ctrl._worker._sync_tp_max_workers, 1)
                self.assertIsNotNone(ctrl._worker._sync_call_tp)

                # Check if the dispatcher still function
                await self._check_if_function_is_ok(host)

            mock_logger.warning.assert_any_call(
                f'{PYTHON_THREADPOOL_THREAD_COUNT} must be an integer')

    async def test_dispatcher_sync_threadpool_below_min_setting(self):
        """Test if the sync threadpool will pick up default value when the
        setting is below minimum
        """

        with patch('azure_functions_worker.dispatcher.logger') as mock_logger:
            # Configure thread pool max worker to an invalid value
            os.environ.update({PYTHON_THREADPOOL_THREAD_COUNT: '0'})
            ctrl = testutils.start_mockhost(
                script_root=self.dispatcher_funcs_dir)

            async with ctrl as host:
                await self._check_if_function_is_ok(host)

                # Ensure the dispatcher sync threadpool should fallback to 1
                self.assertEqual(ctrl._worker._sync_tp_max_workers, 1)
                self.assertIsNotNone(ctrl._worker._sync_call_tp)

                # Check if the dispatcher still function
                await self._check_if_function_is_ok(host)

            mock_logger.warning.assert_any_call(
                f'{PYTHON_THREADPOOL_THREAD_COUNT} must be set to a value '
                'between 1 and 32')

    async def test_dispatcher_sync_threadpool_exceed_max_setting(self):
        """Test if the sync threadpool will pick up default value when the
        setting is above maximum
        """

        with patch('azure_functions_worker.dispatcher.logger') as mock_logger:
            # Configure thread pool max worker to an invalid value
            os.environ.update({PYTHON_THREADPOOL_THREAD_COUNT: '33'})
            ctrl = testutils.start_mockhost(
                script_root=self.dispatcher_funcs_dir)

            async with ctrl as host:
                await self._check_if_function_is_ok(host)

                # Ensure the dispatcher sync threadpool should fallback to 1
                self.assertEqual(ctrl._worker._sync_tp_max_workers, 1)
                self.assertIsNotNone(ctrl._worker._sync_call_tp)

                # Check if the dispatcher still function
                await self._check_if_function_is_ok(host)

            mock_logger.warning.assert_any_call(
                f'{PYTHON_THREADPOOL_THREAD_COUNT} must be set to a value '
                'between 1 and 32')

    async def test_dispatcher_sync_threadpool_in_placeholder(self):
        """Test if the sync threadpool will pick up app setting in placeholder
        mode (Linux Consumption)
        """

        ctrl = testutils.start_mockhost(script_root=self.dispatcher_funcs_dir)

        async with ctrl as host:
            await self._check_if_function_is_ok(host)

            # Reload environment variable on specialization
            await host.reload_environment(environment={
                PYTHON_THREADPOOL_THREAD_COUNT: '3'
            })

            # Ensure the dispatcher sync threadpool should update to 3
            self.assertEqual(ctrl._worker._sync_tp_max_workers, 3)
            self.assertIsNotNone(ctrl._worker._sync_call_tp)

            # Check if the dispatcher still function
            await self._check_if_function_is_ok(host)

    async def test_dispatcher_sync_threadpool_in_placeholder_invalid(self):
        """Test if the sync threadpool will use the default setting when the
        app setting is invalid
        """

        ctrl = testutils.start_mockhost(script_root=self.dispatcher_funcs_dir)

        with patch('azure_functions_worker.dispatcher.logger') as mock_logger:
            async with ctrl as host:
                await self._check_if_function_is_ok(host)

                # Reload environment variable on specialization
                await host.reload_environment(environment={
                    PYTHON_THREADPOOL_THREAD_COUNT: 'invalid'
                })

                # Ensure the dispatcher sync threadpool should fallback to 1
                self.assertEqual(ctrl._worker._sync_tp_max_workers, 1)
                self.assertIsNotNone(ctrl._worker._sync_call_tp)

                # Check if the dispatcher still function
                await self._check_if_function_is_ok(host)

                # Check warning message
                mock_logger.warning.assert_any_call(
                    f'{PYTHON_THREADPOOL_THREAD_COUNT} must be an integer')

    async def test_dispatcher_sync_threadpool_in_placeholder_above_max(self):
        """Test if the sync threadpool will use the default setting when the
        app setting is above maximum
        """

        ctrl = testutils.start_mockhost(script_root=self.dispatcher_funcs_dir)

        with patch('azure_functions_worker.dispatcher.logger') as mock_logger:
            async with ctrl as host:
                await self._check_if_function_is_ok(host)

                # Reload environment variable on specialization
                await host.reload_environment(environment={
                    PYTHON_THREADPOOL_THREAD_COUNT: '33'
                })

                # Ensure the dispatcher sync threadpool should fallback to 1
                self.assertEqual(ctrl._worker._sync_tp_max_workers, 1)
                self.assertIsNotNone(ctrl._worker._sync_call_tp)

                # Check if the dispatcher still function
                await self._check_if_function_is_ok(host)

                # Check warning message
                mock_logger.warning.assert_any_call(
                    f'{PYTHON_THREADPOOL_THREAD_COUNT} must be set to a value '
                    'between 1 and 32')

    async def test_dispatcher_sync_threadpool_in_placeholder_below_min(self):
        """Test if the sync threadpool will use the default setting when the
        app setting is below minimum
        """

        ctrl = testutils.start_mockhost(script_root=self.dispatcher_funcs_dir)

        with patch('azure_functions_worker.dispatcher.logger') as mock_logger:
            async with ctrl as host:
                await self._check_if_function_is_ok(host)

                # Reload environment variable on specialization
                await host.reload_environment(environment={
                    PYTHON_THREADPOOL_THREAD_COUNT: '0'
                })

                # Ensure the dispatcher sync threadpool should fallback to 1
                self.assertEqual(ctrl._worker._sync_tp_max_workers, 1)
                self.assertIsNotNone(ctrl._worker._sync_call_tp)

                # Check if the dispatcher still function
                await self._check_if_function_is_ok(host)

                # Check warning message
                mock_logger.warning.assert_any_call(
                    f'{PYTHON_THREADPOOL_THREAD_COUNT} must be set to a value '
                    'between 1 and 32')

    async def test_sync_invocation_request_log(self):
        ctrl = testutils.start_mockhost(script_root=self.dispatcher_funcs_dir)

        with patch('azure_functions_worker.dispatcher.logger') as mock_logger:
            async with ctrl as host:
                request_id: str = ctrl._worker._request_id
                func_id, invoke_id = await self._check_if_function_is_ok(host)

                mock_logger.info.assert_any_call(
                    'Received FunctionInvocationRequest, '
                    f'request ID: {request_id}, '
                    f'function ID: {func_id}, '
                    f'invocation ID: {invoke_id}, '
                    'function type: sync, '
                    'sync threadpool max workers: 1'
                )

    async def test_async_invocation_request_log(self):
        ctrl = testutils.start_mockhost(script_root=self.dispatcher_funcs_dir)

        with patch('azure_functions_worker.dispatcher.logger') as mock_logger:
            async with ctrl as host:
                request_id: str = ctrl._worker._request_id
                func_id, invoke_id = (
                    await self._check_if_async_function_is_ok(host)
                )

                mock_logger.info.assert_any_call(
                    'Received FunctionInvocationRequest, '
                    f'request ID: {request_id}, '
                    f'function ID: {func_id}, '
                    f'invocation ID: {invoke_id}, '
                    'function type: async'
                )

    async def test_sync_invocation_request_log_threads(self):
        ctrl = testutils.start_mockhost(script_root=self.dispatcher_funcs_dir)
        os.environ.update({PYTHON_THREADPOOL_THREAD_COUNT: '5'})

        with patch('azure_functions_worker.dispatcher.logger') as mock_logger:
            async with ctrl as host:
                request_id: str = ctrl._worker._request_id
                func_id, invoke_id = await self._check_if_function_is_ok(host)

                mock_logger.info.assert_any_call(
                    'Received FunctionInvocationRequest, '
                    f'request ID: {request_id}, '
                    f'function ID: {func_id}, '
                    f'invocation ID: {invoke_id}, '
                    'function type: sync, '
                    'sync threadpool max workers: 5'
                )

    async def test_async_invocation_request_log_threads(self):
        ctrl = testutils.start_mockhost(script_root=self.dispatcher_funcs_dir)
        os.environ.update({PYTHON_THREADPOOL_THREAD_COUNT: '4'})

        with patch('azure_functions_worker.dispatcher.logger') as mock_logger:
            async with ctrl as host:
                request_id: str = ctrl._worker._request_id
                func_id, invoke_id = (
                    await self._check_if_async_function_is_ok(host)
                )

                mock_logger.info.assert_any_call(
                    'Received FunctionInvocationRequest, '
                    f'request ID: {request_id}, '
                    f'function ID: {func_id}, '
                    f'invocation ID: {invoke_id}, '
                    'function type: async'
                )

    async def test_sync_invocation_request_log_in_placeholder_threads(self):
        ctrl = testutils.start_mockhost(script_root=self.dispatcher_funcs_dir)

        with patch('azure_functions_worker.dispatcher.logger') as mock_logger:
            async with ctrl as host:
                await host.reload_environment(environment={
                    PYTHON_THREADPOOL_THREAD_COUNT: '5'
                })

                request_id: str = ctrl._worker._request_id
                func_id, invoke_id = await self._check_if_function_is_ok(host)

                mock_logger.info.assert_any_call(
                    'Received FunctionInvocationRequest, '
                    f'request ID: {request_id}, '
                    f'function ID: {func_id}, '
                    f'invocation ID: {invoke_id}, '
                    'function type: sync, '
                    'sync threadpool max workers: 5'
                )

    async def test_async_invocation_request_log_in_placeholder_threads(self):
        ctrl = testutils.start_mockhost(script_root=self.dispatcher_funcs_dir)

        with patch('azure_functions_worker.dispatcher.logger') as mock_logger:
            async with ctrl as host:
                await host.reload_environment(environment={
                    PYTHON_THREADPOOL_THREAD_COUNT: '5'
                })

                request_id: str = ctrl._worker._request_id
                func_id, invoke_id = (
                    await self._check_if_async_function_is_ok(host)
                )

                mock_logger.info.assert_any_call(
                    'Received FunctionInvocationRequest, '
                    f'request ID: {request_id}, '
                    f'function ID: {func_id}, '
                    f'invocation ID: {invoke_id}, '
                    'function type: async'
                )

    async def _check_if_function_is_ok(self, host) -> Tuple[str, str]:
        # Ensure the function can be properly loaded
        func_id, load_r = await host.load_function('show_context')
        self.assertEqual(load_r.response.function_id, func_id)
        self.assertEqual(load_r.response.result.status,
                         protos.StatusResult.Success)

        # Ensure the function can be properly invoked
        invoke_id, call_r = await host.invoke_function(
            'show_context', [
                protos.ParameterBinding(
                    name='req',
                    data=protos.TypedData(
                        http=protos.RpcHttp(
                            method='GET'
                        )
                    )
                )
            ])
        self.assertIsNotNone(invoke_id)
        self.assertEqual(call_r.response.result.status,
                         protos.StatusResult.Success)

        return (func_id, invoke_id)

    async def _check_if_async_function_is_ok(self, host) -> Tuple[str, str]:
        # Ensure the function can be properly loaded
        func_id, load_r = await host.load_function('show_context_async')
        self.assertEqual(load_r.response.function_id, func_id)
        self.assertEqual(load_r.response.result.status,
                         protos.StatusResult.Success)

        # Ensure the function can be properly invoked
        invoke_id, call_r = await host.invoke_function(
            'show_context_async', [
                protos.ParameterBinding(
                    name='req',
                    data=protos.TypedData(
                        http=protos.RpcHttp(
                            method='GET'
                        )
                    )
                )
            ])
        self.assertIsNotNone(invoke_id)
        self.assertEqual(call_r.response.result.status,
                         protos.StatusResult.Success)

        return (func_id, invoke_id)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import json
import azure.functions as func


async def main(req: func.HttpRequest,
               context: func.Context) -> func.HttpResponse:
    result = {
        'function_directory': context.function_directory,
        'function_name': context.function_name
    }
    return func.HttpResponse(body=json.dumps(result),
                             mimetype='application/json')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as func


# This is an actual EventHub trigger which will convert the event data
# into a storage blob.
def main(event: func.EventHubEvent) -> bytes:
    return event.get_body()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import typing
import json
import azure.functions as func


# This is an actual EventHub trigger which handles Eventhub events in batches.
# It serializes multiple event data into a json and store it into a blob.
def main(events: typing.List[func.EventHubEvent]) -> bytes:
    event_list = []
    for event in events:
        event_dict: typing.Mapping[str, typing.Any] = {
            'body': event.get_body().decode('utf-8'),
            'enqueued_time': event.enqueued_time.isoformat(),
            'partition_key': event.partition_key,
            'sequence_number': event.sequence_number,
            'offset': event.offset,
            'metadata': event.metadata
        }
        event_list.append(event_dict)

    return json.dumps(event_list)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import azure.functions as azf


def main(req: azf.HttpRequest) -> bytes:
    return azf.QueueMessage(body=req.get_body())
